#include <iostream>		/*cout et al*/
#include <libgimp/gimp.h>	/*Access to Gimp internals*/

/*Header for this file*/
#include "plugin.h"
#include "yuvimage.h"
#include "filter.h"

using namespace std;

/*Tell Gimp what our query and run functions are called*/
GimpPlugInInfo PLUG_IN_INFO =
{
	NULL,
	NULL,
	query,
	run
};


/*Black pre-processor magic...Needed to make this into a Gimp plugin.*/
MAIN()


/*Function that tells Gimp what our plugin can do, and what parameters it takes*/
static void query ()
{
	/*Tell Gimp what parameters this filter takes: a type, a name, and a description*/
	static GimpParamDef args[] =
	{
		{
			GIMP_PDB_INT32,
			"run-mode",
			"Run mode"
		},
		{
			GIMP_PDB_IMAGE,
			"image",
			"Input image"
		},
		{
			GIMP_PDB_DRAWABLE,
			"drawable",
			"Input drawable"
		}
	};
	
	/*Install it in Gimp's database*/
	gimp_install_procedure (
		"plug-in-karl",
		"Karl\'s Recolorizing Filter",
		"Regenerates the colours for an image based on quick colour scribbles",
		"Karl Erlandsen",
		"Copyright Karl Erlandsen",
		"2006",
		"_Karls Funky Filter",
		"RGB*, GRAY*",
		GIMP_PLUGIN,
		G_N_ELEMENTS (args), 0,
		args, NULL);
	
	/*Stick a menu entry in for it*/
	gimp_plugin_menu_register ("plug-in-karl",
				"<Image>/Filters/");
}

/*Function that actually performs our filtering operation*/
static void run (const gchar      *name,
		 gint              nparams,
		 const GimpParam  *param,
		 gint             *nreturn_vals,
		 GimpParam       **return_vals)
{
	/*Return values*/
        static GimpParam  values[1];
        GimpPDBStatusType status = GIMP_PDB_SUCCESS;
	/*Inputs into our filter*/
        GimpDrawable     *drawable;
	
	/*Where the user has asked us to filter*/
	gint         fromX, fromY, toX, toY;
	/*Pixel regions - use these to get and set data in Gimp*/
	GimpPixelRgn rgn_in, rgn_out;
	

        /* Setting return values - 1 return value, stored in "values". Note that return values
	are not returned as is the norm for C/C++ programs, but are instead sent back through
	these variables */
        *nreturn_vals = 1;
        *return_vals  = values;

	/*Making our "values" return value the correct type, and storing a successful value in it*/
        values[0].type = GIMP_PDB_STATUS;
        values[0].data.d_status = status;

	/*  Get the specified drawable  */
	drawable = gimp_drawable_get (param[2].data.d_drawable);
	
	/*Find what coordinates the user has selected for us to filter (this defaults to the
	whole image if nothing is selected)*/
	gimp_drawable_mask_bounds (drawable->drawable_id,
				   &fromX, &fromY,
				   &toX, &toY);
	/*Find the number of channels we're dealing with*/
	gint numChannels = gimp_drawable_bpp (drawable->drawable_id);
	
	int width = toX - fromX;
	int height = toY - fromY;
	
	/*Initialise our drawables so we can get and set data in Gimp's image*/
	gimp_pixel_rgn_init (&rgn_in,
			      drawable,
			      fromX, fromY,
			      width, height, 
			      FALSE, FALSE);
	gimp_pixel_rgn_init (&rgn_out,
			      drawable,
			      fromX, fromY,
			      width, height, 
			      TRUE, TRUE);
	
	
	/*Create memory to store our image's RGB values in for solving*/
	yuvimage input(width, height);
	yuvimage marked(width, height);
	yuvimage filtered(width, height);
	
	/*Get a buffer to translate between my code and Gimp's internals*/
	int size = width * height * numChannels;
	guchar *buffer = new guchar[size];
	
	/*Copy image data into it*/
	gimp_pixel_rgn_get_rect  (&rgn_in,
				       buffer,
				       fromX,
				       fromY,
				       width,
				       height);
	
	/*Copy data from our buffer to my image*/
	int i=0;
	for(int y=0; y!=height; ++y)
	{
		for(int x=0; x!=width; ++x, i+=numChannels)
		{
			int pixelOffset = x + (y * width);
			
			/*Make our input image greyscale*/
			input.r[pixelOffset] = ((double)buffer[i]) / 255.0;
			input.g[pixelOffset] = ((double)buffer[i]) / 255.0;
			input.b[pixelOffset] = ((double)buffer[i]) / 255.0;
			
			/*Make our marked image a colour copy*/
			marked.r[pixelOffset] = ((double)buffer[i])/255.0;
			marked.g[pixelOffset] = ((double)buffer[i+1])/255.0;
			marked.b[pixelOffset] = ((double)buffer[i+2])/255.0;
		}
	}
	
	/*Give some progress feedback*/
	gimp_progress_init ("Colorizing...");
	cout << "Colorizing from " << fromX << "," << fromY << " to " << toX << "," << toY << endl;
	cout << numChannels << " channels detected.\n";

	/*Perform our filter operation*/
        filter(input, marked, filtered, 2);
	
	filtered.YUVtoRGB();
	
	i=0;
	for(int y=0; y!=height; ++y)
	{
		for(int x=0; x!=width; ++x, i+=numChannels)
		{
			int pixelOffset = x + (y * width);
			
			buffer[i] = (guchar)(filtered.r[pixelOffset] * 255.0);
			buffer[i+1] = (guchar)(filtered.g[pixelOffset] * 255.0);
			buffer[i+2] = (guchar)(filtered.b[pixelOffset] * 255.0);
		}
	}
	
	
	/*Copy our modified pixels back into Gimp's image*/
	gimp_pixel_rgn_set_rect  (&rgn_out,
				buffer,
				fromX,
				fromY,
				width,
				height);
	
	/*Free our buffer since it's now no longer needed*/
	delete [] buffer;

	/*Tell Gimp to update its image*/

	gimp_drawable_flush(drawable);
	gimp_drawable_merge_shadow (drawable->drawable_id, TRUE);
	gimp_drawable_update (drawable->drawable_id,
			      fromX, fromY,
			      width, height);
	
	gimp_displays_flush ();
        gimp_drawable_detach (drawable);
}
