/***************************************************************************
*   Copyright (C) 2005 by Karl Erlandsen #c1198379                         *
*   c1198379@bmth.ac.uk                                                    *
*                                                                          *
* This file contains code for the class "yuvimage", which loads in image   *
* data through ImageMagick and provides the user with an interface to      *
* access the data as either YUV or RGB double floating point arrays. This  *
* class is used to load, store, and finally display processed data through *
* an OpenGL texture.                                                       *
*                                                                          *
* Not for use without the authors permission				   *
***************************************************************************/

#include <iostream>	//cout, cerr et al
#include <stdlib.h>	//exit
#include <assert.h>     //assert
#include <Magick++.h>	//ImageMagick image processing routines
#include <GL/gl.h>	//For OpenGL texture generation
#include <GL/glu.h>	//For OpenGL texture generation

#include "yuvimage.h"	//Header for this class

using namespace std;

yuvimage::yuvimage()
{
	/*Give everything sane default values*/
	initMembers();

	return;
}


yuvimage::yuvimage(const int width_, const int height_)
{
	/*Give everything sane default values*/
	initMembers();

	/*Set our size members*/
	width = width_;
	height = height_;
	
	/*Get memory for pixmap.*/
	r = new double [width * height];
	g = new double [width * height];
	b = new double [width * height];
	
	assert(r != NULL);
	assert(g != NULL);
	assert(b != NULL);
	
	/*Initialise pixmap to be black*/
	for(int i = 0; i != width * height; ++i)
		r[i] = g[i] = b[i] = 0.0;
	
	/*Set other pointers to NULL so we know we're not using them any more*/
	y = u = v = (double*)NULL;
	
	return;
}


/*constructor that loads the given image file into our object as RGB[0.0-1.0] data*/
yuvimage::yuvimage(const string Filename)
{
	/*Give everything sane default values*/
	initMembers();
	
	loadImage(Filename);
	
	return;
}


/*Destructor*/
yuvimage::~yuvimage()
{
	/*free dynamically allocated pixel memory before it's destroyed*/
	freePixmap();
	
	/*Free our OpenGL texture, if it exists*/
	if(texture != 0)
		glDeleteTextures(1, &texture);
	
	return;
}

/*function that tries to load the given image file into our object as RGB[0.0 - 1.0] data.
 *Note that at the moment if it fails, the whole program will exit.*/
void yuvimage::loadImage(const string Filename)
{
	/*Our imagemagick image*/
	Magick::Image input;
	
	/*Load in a texture, exit if it fails*/
	try
	{
		input.read(Filename);
	}
	catch (Magick::Exception &error_ )
	{
		cerr << "ERROR: error while loading image: " << error_.what();
		cerr << endl;
		exit(EXIT_FAILURE);
	}
	
	/*Get image dimensions*/
	width = input.columns();
	height = input.rows();
	cout << "INFO: Read in image with size " << width << " x " << height << endl;
	
	/*Free our pixmap if it's already been allocated*/
	freePixmap();
	
	/*Allocate memory for our pixmap, and copy the data into it. We must read in as RGB because
	ImageMagick does not allow us to read in directly as YUV. */
	r = new double[width * height];
	g = new double[width * height];
	b = new double[width * height];
	
	assert(r != NULL);
	assert(g != NULL);
	assert(b != NULL);
	
	/*Load our ImageMagick data into R, G and B channels.*/
	input.write(0, 0, width, height, "R", Magick::DoublePixel, r);
	input.write(0, 0, width, height, "G", Magick::DoublePixel, g);
	input.write(0, 0, width, height, "B", Magick::DoublePixel, b);
	
	return;
}


/*Function to convert the image's data from RGB to YUV*/
void yuvimage::RGBtoYUV()
{
	/*Dont do anything if our image has already been converted to YUV*/
	if(!isRGB)
		return;
	
	/*Swap pointers about, so we don't need to waste time with memory allocation*/
	y = r;
	u = g;
	v = b;
	
	/*Convert our pixmap from RGB to YUV*/
	for(int i=0; i!=(height * width); ++i)
	{
		/*Conversion details from http://www.answers.com/topic/yuv-rgb-conversion-formulas*/
		double newY = (	  ( 0.299 * r[i] )
					+ ( 0.587 * g[i] )
					+ ( 0.114 * b[i] )  );
	
		double newU = (	  (-0.147 * r[i] )
					+ (-0.289 * g[i] )
					+ ( 0.436 * b[i] )  );
	
		double newV = (	  ( 0.615 * r[i] )
					+ (-0.515 * g[i] )
					+ (-0.100 * b[i] )  );
		
		y[i] = newY;
		u[i] = newU;
		v[i] = newV;
	}
	
	/*Set RGB pointers to NULL so we know we're not using them any more*/
	r = g = b = (double*)NULL;
	
	/*Set a flag so that we know what state our pixmap is in*/
	isRGB=0;
	
	return;
}


/*Function to convert the image's data from YUV back to RGB. It is lossy, but only by a value of about
0.01 or so per sample, and it seems to hold up well under repeated use*/
void yuvimage::YUVtoRGB()
{
	/*Dont do anything if our image has already been converted to RGB*/
	if(isRGB)
		return;
	
	/*Swap pointers about, so we don't need to waste time with memory allocation*/
	r = y;
	g = u;
	b = v;
	
	/*Convert our pixmap from YUV to RGB*/
	for(int i=0; i!=(height * width); ++i)
	{
		/*Conversion details from http://www.answers.com/topic/yuv-rgb-conversion-formulas*/
		double newR = ( y[i] + (1.140 * v[i]) );
	
		double newG = ( y[i] - (0.395 * u[i]) - (0.581 * v[i]) );
	
		double newB = ( y[i] + (2.032 * u[i]) );
		
		/*We clamp values just to be on the safe side, as this function has been known to produce
		negative values of blue. This would certainly be possible with eg. y=0.0 and u=-0.5. Needless
		to say, OpenGL gets rather unhappy about negative colour values, and wraps them round to the
		positive end of the scale. This can lead to eg. big blue splodges in the middle of your image
		for no apparent reason.*/
		if(newR < 0.0)
			newR = 0.0;
		if(newR > 1.0)
			newR = 1.0;
		if(newG < 0.0)
			newG = 0.0;
		if(newG > 1.0)
			newG = 1.0;
		if(newB < 0.0)
			newB = 0.0;
		if(newB > 1.0)
			newB = 1.0;
	
		r[i] = newR;
		g[i] = newG;
		b[i] = newB;
		
	}
	
	/*Set YUV pointers to NULL*/
	y = u = v = (double*)NULL;
	
	/*Set a flag so that we know what state our pixmap is in*/
	isRGB = 1;
	
	return;
}


/*Function to convert our image data from floating point YUV/RGB into unsigned char
RGB data from 0 to 255, and convert this to an OpenGL texture*/
void yuvimage::makeDrawData()
{
	int flag = 0;
	int size = (height * width);
	unsigned char *drawData = NULL;
	
	/*Convert our image if this is necessary*/
	if(!isRGB)
	{
		YUVtoRGB();
		flag = 1;
	}
	
	/*Get data for drawing data*/
	drawData = new unsigned char[size * 3];
	assert(drawData != NULL);
	
	int n = 0;
	//int first = 1;
	/*Convert to range 0-255*/
	for(int i=0; i!=size; ++i)
	{
		drawData[n] = (unsigned char)(r[i] * 255.0);
		drawData[n+1] = (unsigned char)(g[i] * 255.0);
		drawData[n+2] = (unsigned char)(b[i] * 255.0);
		n+=3;
	}
	
	
	/*Make an OpenGL texture*/
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, width,
			  height, GL_RGB, GL_UNSIGNED_BYTE, 
			  drawData);
	assert( !glGetError() );
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	
	/*No longer need drawData so we free it*/
	delete [] drawData;
	drawData = NULL;
	
	/*Convert our image back to YUV if it was so before*/
	if(flag)
		RGBtoYUV();
	
	return;
}


/*Simple function to initialise all class members to 0 or other sane defaults*/
void yuvimage::initMembers()
{
	height = 0;
	width = 0;
	isRGB=1;
	r = NULL;
	g = NULL;
	b = NULL;
	y = NULL;
	u = NULL;
	v = NULL;
	texture = 0;
}


/*This function frees all allocated pixel memory, and set the associated pointer to NULL afterwards*/
void yuvimage::freePixmap()
{
	if(r != NULL)
	{
		delete [] r;
		r = (double*)NULL;
	}
	if(g != NULL)
	{
		delete [] g;
		g = (double*)NULL;
	}
	if(b != NULL)
	{
		b = (double*)NULL;
		delete [] b;
	}
	if(y != NULL)
	{
		y = (double*)NULL;
		delete [] y;
	}
	if(u != NULL)
	{
		u = (double*)NULL;
		delete [] u;
	}
	if(v != NULL)
	{
		v = (double*)NULL;
		delete [] v;
	}
	
	return;
}
