/***************************************************************************
 *   Copyright (C) 2005 by Karl Erlandsen #c1198379                        *
 *   c1198379@bmth.ac.uk                                                   *
 *                                                                         *
 * Innovations programming project, written February 2006. This is an      *
 * implementation of the SIGGRAPH 2004 paper "Colorization Using           *
 * Optimization", available from                                           *
 * "http://www.cs.huji.ac.il/~yweiss/Colorization/"                        *
 *                                                                         *
 * Not for use without the authors permission                              *
 *                                                                         *
 * You are also hereby warned that this code is, by the author's own       *
 * admission, horrific. It was written in a rush so that it would be ready *
 * quickly, not so that it would be fast, efficient, easy to read or       *
 * "beautiful" (for those people who claim code can be beautiful).         *
 * It's big, it's unweildy, and it just doesn't care. Use at your own risk.*
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream> 	//cout, cerr et al
#include <stdlib.h> 	//exit
#include "SDL.h"	//standard SDL stuff, for windowing
#include <GL/gl.h>	//standard OpenGL stuff, for drawing our image to screen
#include <GL/glu.h>	//GLU library, makes OpenGL easier to work with


/*Local headers*/
#include "yuvimage.h"	//YUVImage class, for loading in an image and converting to/from RGB and YUV
#include "filter.h"	//filter function


/*Maximum size our window can have, just to keep things sane*/
#define MAX_WINDOW_WIDTH (1280)
#define MAX_WINDOW_HEIGHT (1024)


using namespace std;

//The current image that we're drawing to screen - could be the black and white image, the marked image,
//the filtered image, or a reference colour image
yuvimage *current = NULL;

void initSDL(const int width, const int height);
void initGL(const int width, const int height);
void redraw();


int main(int argc, char *argv[])
{
	SDL_Event event;
	bool done = false;
	int width = 0, height = 0;
	/*FIXME: Change program to take arguments as image filename input instead*/

	if(argc != 4)
	{
		cout << "\n\nInnovations programming project, by Karl \"Linux Karl\" Erlandsen, 28/2/2006.\n";
		cout << "Based on the SIGGRAPH 2004 paper \"Colorization Using Optimization\"(sic), from\n";
		cout << "http://www.cs.huji.ac.il/~yweiss/Colorization/\n\n";
		cout << "Usage: \"filter BlackAndWhiteImage MarkedImage ReferenceColourImage\"\n";
		cout << "where each image may be any format readable by ImageMagick, up to " << MAX_WINDOW_WIDTH << " x ";
		cout << MAX_WINDOW_HEIGHT << ".\n";
		cout <<"After a short pause a window will appear - press '1' to show the black and white image,\n";
		cout <<"press '2' to show the marked image, press '3' to see the result of the filter operation,\n";
		cout <<"and press '4' to compare the results with the reference colour image.\n\n\n";
		exit(EXIT_FAILURE);
	}
	
	/*Create filenames from input arguments*/
	string inputImagePath(argv[1]);
	string markedImagePath(argv[2]);
	string colourImagePath(argv[3]);
	
	/*Load in images*/
	yuvimage input(inputImagePath);
	yuvimage marked(markedImagePath);
	yuvimage original(colourImagePath);
	yuvimage filtered(input.width, input.height);
	
	/*Check our two input images are the same size*/
	if(input.width != marked.width || input.height != marked.height)
	{
		cerr << "ERROR: input and marked images are of different size! Exiting.\n";
		exit(EXIT_FAILURE);
	}
	
	/*Set our window size to the same size as our image.
	We clamp our size values if they're stupidly big*/
	width = input.width;
	height = input.height;
	if(width > MAX_WINDOW_WIDTH)
		width = MAX_WINDOW_WIDTH;
	if(height > MAX_WINDOW_HEIGHT)
		height = MAX_WINDOW_HEIGHT;
	
	/* Initialise SDL/OpenGL window to the size of our image*/
	initSDL(width, height);
	initGL(width, height);
	
	/*Run our image filter, with a sample size of 1.*/
	filter(input, marked, filtered, 3);
	
	/*Convert our image data into OpenGL textures*/
	input.makeDrawData();
	marked.makeDrawData();
	filtered.makeDrawData();
	original.makeDrawData();
	
	/*Set which image to draw to screen - filtered image by default*/
	current = &filtered;
		
	/*Redraw once to clear the screen*/
	redraw();
		
	/* Event loop - draw images to screen until the user quits*/
	/*Wait for an event to occur*/
	while(SDL_WaitEvent(&event) && done != true)
	{
		switch(event.type)
		{
			/*If the user's just released a keyboard key, we do stuff*/
			case SDL_KEYUP:
			switch(event.key.keysym.sym)
			{
				/*'1' key - we show our B&W input image*/
				case SDLK_1:
				current = &input;
				break;
				/*'2' key - we show our marked image*/
				case SDLK_2:
				current = &marked;
				break;
				/*'3' key - we show our filtered image*/
				case SDLK_3:
				current = &filtered;
				break;
				/*'4' key - we show the colour image*/
				case SDLK_4:
				current = &original;
				break;
			}
			/*If the user's just exposed part of our window we should redraw*/
			case SDL_VIDEOEXPOSE:
			redraw();
			break;
			/*If the user wants to quit, we prepare to exit*/
			case SDL_QUIT:
			done = true;
			break;
		}
	}
	
	cout << "INFO: All done, exiting normally.\n";
	exit(EXIT_SUCCESS);
}



/*Function to initialise SDL window/event management*/
void initSDL(const int width, const int height)
{
	/* Information about the current video settings. */
	const SDL_VideoInfo* info = NULL;
	/* Flags we will pass into SDL_SetVideoMode. */
	int flags = 0;
	/*Window parameters*/
	int bpp = 0;
	
	
	/*Initialise SDL, quit if this fails*/
	if(SDL_Init(SDL_INIT_VIDEO) < 0)
	{
		cerr <<"ERROR: Could not initialise SDL: " << SDL_GetError() << endl;
		SDL_Quit();
		exit(EXIT_FAILURE);
	}
	
	/*In future whenever we call exit, we want SDL_Quit to be run first*/
	atexit(SDL_Quit);
	
	/* Get display information, quit if this fails */
	info = SDL_GetVideoInfo( );
	if( info == NULL )
	{
		cerr << "ERROR: Video query failed: " << SDL_GetError() << endl;
		exit(EXIT_FAILURE);
	}
	
	/* Set our width/height to 640/480 (you would
	* of course let the user decide this in a normal
	* app). We get the bpp we will request from
	* the display. On X11, VidMode can't change
	* resolution, so this is probably being overly
	* safe. Under Win32, ChangeDisplaySettings
	* can change the bpp. */
	
	
	bpp = info->vfmt->BitsPerPixel;
	
	
	/* Now, we want to setup our requested
	* window attributes for our OpenGL window.
	* We want *at least* 8 bits of red, green
	* and blue. We also want at least a 16-bit
	* depth buffer.
	*
	* The last thing we do is request a double
	* buffered window. '1' turns on double
	* buffering, '0' turns it off.
	*
	* Note that we do not use SDL_DOUBLEBUF in
	* the flags to SDL_SetVideoMode. That does
	* not affect the GL attribute state, only
	* the standard 2D blitting setup. */
	
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
	
	/* Request that SDL provide us with an OpenGL window.
	* If we wanted other funky stuff like fullscreen, we'd
	* do that here. */
	flags = SDL_OPENGL;
	
	/* Set the video mode, exit if it fails. This could happen
	* for a variety of reasons, including DISPLAY not being set,
	* the specified resolution not being available, etc. */
	if( SDL_SetVideoMode( width, height, bpp, flags ) == 0 )
	{
		cerr << "ERROR: Video mode set failed: " << SDL_GetError( ) << endl;
		exit( EXIT_FAILURE );
	}
	
	return;
}
	
void initGL(const int width, const int height)
{
	/*Set the clear color.*/
	glClearColor( 0, 0, 0, 0 );
	glClearDepth(1.0);
	
	/*Setup our viewport.*/
	glViewport( 0, 0, width, height );
	
	/*Set up our initial camera angles and attributes*/
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity( );
	glOrtho( 0.0, 1.0,	/*Left, right*/
		 0.0, 1.0,	/*Bottom, top*/
		-10.0, 10.0);	/*Near, far*/
		
	/*Set up texturing*/
	glEnable(GL_TEXTURE_2D);
	
	return;
}
	
void redraw()
{
	/*Clear the background*/
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	/*Prepare for texture drawing*/
	glBindTexture(GL_TEXTURE_2D, current->texture);
	
	/*Redraw a textured square*/
	glColor3f(1.0, 1.0, 1.0);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 1.0);
		glVertex3f(0.0, 0.0, 5.0);
			
		glTexCoord2f(1.0, 1.0);
		glVertex3f(1.0, 0.0, 5.0);
			
		glTexCoord2f(1.0, 0.0);
		glVertex3f(1.0, 1.0, 5.0);
			
		glTexCoord2f(0.0, 0.0);
		glVertex3f(0.0, 1.0, 5.0);
	glEnd();
	glFlush();
	
	/*Swap buffers*/
	SDL_GL_SwapBuffers();
	
	return;
}

