/***************************************************************************
 *   Copyright (C) 2005 by Karl Erlandsen #c1198379                        *
 *   c1198379@bmth.ac.uk                                                   *
 *                                                                         *
 * This file contains the filtering code that actually performs the        *
 * recolorizing operation.                                                 *
 *                                                                         *
 * Not for use without the authors permission				   *
 ***************************************************************************/

#include <iostream>		//cout et al
#include <math.h>		//exp
#include <assert.h>		//assert

extern "C" {
#include <umfpack/umfpack.h>	//UMFPack, for linear system solving. It's a C library, so we have
}				//to wrap it in extern "C" to stop linking errors

#include "yuvimage.h"		//YUVImage class, which we use to access image data
#include "tripletmatrix.h"	//Matrix class we store weighting values in before
				//solving with UMFPack

#include "filter.h"		//This file

using namespace std;

/*Local function declarations*/
int isRoughlyEqual(double value1, double value2);
void calculateWeighting(yuvimage &input, int posX, int posY, tripletMatrix &matrix, unsigned int sampleSize);


/*The threshold of when we count floating point numbers as equalling each other*/
#define REALLY_SMALL (0.000002)


/*Simple function I knocked up because I got tired of floating point rounding errors.
This function returns 1 if two floating point numbers are near enough to each other
that it makes no real difference, else it returns 0*/
int isRoughlyEqual(double value1, double value2)
{
    double difference = value1 - value2;
    if(difference < REALLY_SMALL && difference > -REALLY_SMALL )
        return 1;
    else
        return 0;
}


/*Our filter function. It takes 3 YUV image inputs - a black and white image ("input"),
an image with some pixel colours marked by the user ("marked") and an image to write the
results of the filtering operation to ("filtered"). It uses collects samples in a radius
of "sampleSize" around each pixel - higher equals better results, but much more calculation time - 
1 is a good defult value*/
void filter(yuvimage &input, yuvimage &marked, yuvimage &filtered, unsigned int sampleSize)
{
	cout << "INFO: Beginning filter operation.\n";
	cout.flush();
	
	/*Calculate how many samples we'll be taking per pixel*/
	int sampleRadius = (2*sampleSize) + 1;
	int numSamples = (sampleRadius * sampleRadius);
	
	/*Matrix we store our weight calculations in (in both triplet and compressed column notation)*/
	tripletMatrix matrix(input.width, input.height, numSamples);
	
	/*Convert our images to YUV if they aren't already*/
	input.RGBtoYUV();
	marked.RGBtoYUV();
	filtered.RGBtoYUV();
	
	/*Set all non-user marked U and V samples to equal 0.0 - this is important for linear system
	solving*/
	for (int y = 0; y != input.height; y++)
	{
		for (int x = 0; x != input.width; x++)
		{
			int pixelOffset = x + (y*input.width);
 			
			/*If the user hasn't marked this pixel then we set that colour to 0,
			and set a flag value so we know to run the algorithm on this pixel*/
			if( (isRoughlyEqual(input.y[pixelOffset], marked.y[pixelOffset]) )
			&&  isRoughlyEqual(input.u[pixelOffset], marked.u[pixelOffset])
			&&  isRoughlyEqual(input.v[pixelOffset], marked.v[pixelOffset])  )
			{
				marked.u[pixelOffset] = 0.0;
				marked.v[pixelOffset] = 0.0;
			}
		}
	}
	
	/*Itterate through each pixel in the image to calculate several weighting samples*/
	for (int y = 0; y != input.height; y++)
	{
		for (int x = 0; x != input.width; x++)
		{
			int pixelOffset = x + (y * input.width);
			
			/*Calculate the weighting of all neighbouring pixels in a window round this pixel, and add
			them to the weight matrix. Only do so if it hasn't been marked with a colour by the user,
			however.*/
			if( (isRoughlyEqual(input.y[pixelOffset], marked.y[pixelOffset]) )
			&&  isRoughlyEqual(input.u[pixelOffset], marked.u[pixelOffset])
			&&  isRoughlyEqual(input.v[pixelOffset], marked.v[pixelOffset])  )
				calculateWeighting(input, x, y, matrix, sampleSize);
			
			/*Make an element of the matrix set the pixel to have a perfect weighting relative to itself
			(not sure how this plays with the algorithm, since according to the paper the weightings for
			any pixel are *meant* to sum to 1...)*/
			matrix.addTriplet(1.0, pixelOffset, pixelOffset);
		}
	}

	/*UMFPack solving handles*/
	void *symbolic, *numeric;

	/*Convert our matrix to compressed column notation so that UMFPack can use it*/
	matrix.convertToCompressedColumn();

	/*Perform symbolic and numerical analysis of the column form of our matrix*/
	int result;
	cout << "Beginning symbolic analysis." << endl;
	cout.flush();
	result = umfpack_di_symbolic(	input.height * input.width,
				      	input.height * input.width,
					matrix.columnP,
					matrix.columnI,
					matrix.columnX,
					&symbolic,
					NULL, NULL);
	if(result == 0)
	{
		cout << "Symbolic analysis successful." << endl;
		cout.flush();
	}
	else
	{
		cerr << "WARNING: symbolic analysis returned non-success value " << result << endl;
		cerr.flush();
	}
	
	cout << "Beginning numeric analysis." << endl;
	cout.flush();
	result = umfpack_di_numeric(matrix.columnP,
				    matrix.columnI,
				    matrix.columnX,
				    symbolic, &numeric,
				    NULL, NULL);
	if(result == 0)
	{
		cout << "Numeric analysis successful." << endl;
		cout.flush();
	}
	else
	{
		cerr << "WARNING: numeric analysis returned non-success value " << result << endl;
		cerr.flush();
	}
	
	umfpack_di_free_symbolic(&symbolic);
	
	/*Solve the linear system for the U channel*/
	cout << "Beginning solve for U channel." << endl;
	cout.flush();
	result = umfpack_di_solve(UMFPACK_A,
				  matrix.columnP,
				  matrix.columnI,
				  matrix.columnX,
				  filtered.u,
				  marked.u,
				  numeric,
				  NULL, NULL);
	if(result == 0)
	{
		cout << "U channel successfully solved" << endl;
		cout.flush();
	}
	else
	{
		cerr << "WARNING: U solve returned non-success value " << result << endl;
		cout.flush();
	}
	
	/*Solve the linear system for the V channel*/
	cout << "Beginning solve for V channel." << endl;
	cout.flush();
	result = umfpack_di_solve(UMFPACK_A,
				  matrix.columnP,
				  matrix.columnI,
				  matrix.columnX,
				  filtered.v,
				  marked.v,
				  numeric,
				  NULL, NULL);
	if(result == 0)
	{
		cout << "V channel successfully solved" << endl;
		cout.flush();
	}
	else
	{
		cerr << "WARNING: V solve returned non-success value " << result << endl;
		cerr.flush();
	}

	umfpack_di_free_numeric(&numeric);

	/*Finally, copy Y channel values from our input image into our resulting filtered image.*/
	for (int y = 0; y != input.height; y++)
	{
		for (int x = 0; x != input.width; x++)
		{
			int pixelOffset = x + (y * input.width);
			
			filtered.y[pixelOffset] = input.y[pixelOffset];
		}
	}
	
	return;
}


/*This function is run for each pixel that is to be filtered. It calculates weighting values according
  to the affinity function described in the SIGGRAPH paper this is based on (the function in question
  is the squared difference between intensities), and then stores them in "matrix" so that we can run
  linear system solving on them later*/
void calculateWeighting(yuvimage &input, int posX, int posY, tripletMatrix &matrix, unsigned int sampleSize)
{
	/*How far through our colour buffer arrays we are - cached here for speed*/
	int pixelOffset = posX + (posY * input.width);
	
	/*Calculate how many samples we'll be taking per pixel*/
	int sampleRadius = (2*sampleSize) + 1;
	int numSamples = (sampleRadius * sampleRadius);
	
	/*Calculate the bounds of our sampling window*/
	int fromY = posY - (signed int)sampleSize;
	int toY = posY + (signed int)sampleSize;
	int fromX = posX - (signed int)sampleSize;
	int toX = posX + (signed int)sampleSize;
	
	/*Clip them to the image size*/
	if(fromY < 0)
		fromY = 0;
	if(toY > input.height-1)
		toY = input.height - 1;
	if(fromX < 0)
		fromX = 0;
	if(toX > input.width-1)
		toX = input.width - 1;
	
	/*Loop through our samples, and use the pixel intensity to calculating variance.
	The algorithm was taken (butchered) from http://en.wikipedia.org/wiki/Variance/Algorithm,
	accessed 24/1/2006*/
	double sumOfNSquared = 0.0;
	double sumOfN = 0.0;
	for (int y = fromY; y <= toY; ++y)
	{
		for (int x = fromX; x <= toX; ++x)
		{
			/*Get intensity values, add em to sum of n and sum of n squared*/
			int sampleOffset = x + (y * input.width);
			sumOfN += input.y[sampleOffset];
			sumOfNSquared += input.y[sampleOffset] * input.y[sampleOffset];
		}
	}
	double mean = sumOfN/numSamples;
	double variance = (sumOfNSquared - sumOfN * mean) / (numSamples - 1);
	
	/*stop the variance from getting too small. This is to stop weights spiking to infinity
	when we divide by variance a few lines further on*/
	if(variance < REALLY_SMALL)
		variance = REALLY_SMALL;
	
	/*Calculate weighting values (squared difference of intensities) and store them in "weights". Also keep a sum of
	weights in sumOfWeights, so we can scale all weights down so they sum to 1 in a moment.*/
	double *weights = new double[numSamples];
	assert(weights != NULL);
	double sumOfWeights = 0.0;
	int i=0;
	/*Loop through each sample inside the window and calculate weighting values*/
	for(int y = fromY; y <= toY; ++y)
	{
		for(int x = fromX; x <= toX; ++x)
		{
			/*Skip the central pixel in the window - we're only interested in those around it*/
			if(x == posX && y == posY)
				continue;
			
			/*How far through each colour array the current pixel is - cached here for speed*/
			int sampleOffset = x + (y * input.width);
			
			/*Calculate weighting value*/
			weights[i] = (input.y[sampleOffset] - input.y[pixelOffset]);
			weights[i] *= weights[i];
			weights[i] = exp( -weights[i] / variance );
			
			/*Also calculate sum of weights, so we can scale them all down to sum to 1 later*/
			sumOfWeights += weights[i];
			++i;
		}
	}
	
	/*Loop through our sample one last time, scale our weights down, and add them to our matrix*/
	i = 0;
	for(int y = fromY; y <= toY; ++y)
	{
		for(int x = fromX; x <= toX; ++x)
		{
			/*Skip the central pixel in the window - we're only interested in those around it*/
			if(x == posX && y == posY)
				continue;
			
			int sampleOffset = x + (y * input.width);
			
			/*Add our values to our matrix*/
			matrix.addTriplet( -weights[i]/sumOfWeights, pixelOffset, sampleOffset);
			++i;
		}
	}
	
	delete [] weights;
	
	return;
}
