/***************************************************************************
*   Copyright (C) 2005 by Karl Erlandsen #c1198379                        *
*   c1198379@bmth.ac.uk                                                   *
*                                                                         *
* This file contains code for a class to hold a matrix that begins in     *
* triplet form. Triplets of values can be added to it, and then once all  *
* the data has been added it can be converted to a compressed column form *
* matrix instead. Note that once it has been converted, it can no         *
* longer be edited. This class is used to store weighting values for the  *
* image to be filtered, before conversion and use with UMFPack.           *
*                                                                         *
* Not for use without the authors permission				  *
***************************************************************************/

#include <iostream>		//cout et al
#include <assert.h>		//assert

extern "C" {			//UMFPack, for linear system solving and matrix
#include <umfpack/umfpack.h>	//conversion. It's a C library, so we have
}				//to wrap it in extern "C" to stop linking errors

#include "tripletmatrix.h"	//This file

using namespace std;

tripletMatrix::tripletMatrix(int width_, int height_, int windowSize_)
{
	/*Initialise members to sensible defaults*/
	width = width_;
	height = height_;
	windowSize = windowSize_;
	n = 0;
	columnI = columnP = (int*)NULL;
	columnX = (double*)NULL;
	isTriplet = 1;
	
	/*Allocate memory for a triplet-form matrix of the appropriate size*/
	tripletX = new double [width * height * windowSize];
	tripletI = new int [width * height * windowSize];
	tripletJ = new int [width * height * windowSize];
	
	assert(tripletX != NULL);
	assert(tripletI != NULL);
	assert(tripletJ != NULL);
	
	/*Initialise our matrix elements to all be zero - by default no pixels affect any other*/
	for(int i=0; i!=(width * height * windowSize); ++i)
	{
		tripletX[i] = 0.0;
		tripletI[i] = tripletJ[i] = 0;
	}
	
	
	return;
}


tripletMatrix::~tripletMatrix()
{
	/*Free all memory that we were using*/
	if(tripletX != NULL)
		delete [] tripletX;
	if(tripletI != NULL)
		delete [] tripletI;
	if(tripletJ != NULL)
		delete [] tripletJ;
	if(columnP != NULL)
		delete [] columnP;
	if(columnI != NULL)
		delete [] columnI;
	if(columnX != NULL)
		delete [] columnX;
	
	return;
}

/*Function to add a triplet of values to our matrix when it's in triplet form*/
void tripletMatrix::addTriplet(double value, int i, int j)
{
	
	/*If this object is still in "Triplet mode"*/
	if(isTriplet)
	{
		/*Add the values onto the end of the matrix*/
		tripletX[n] = value;
		tripletI[n] = i;
		tripletJ[n] = j;
		++n;
	}
	else
		cerr << "ERROR: cannot add triplet to a matrix which has already been converted.\n";
	
	return;
}

/*Function to convert our matrix from triplet form into compressed column notation*/
void tripletMatrix::convertToCompressedColumn()
{
	/*Don't do anything if this matrix has already been converted*/
	if(!isTriplet)
		return;
	
	/*Allocate memory for compressed column form matrix*/
	columnP = new int [width * height];
	columnI = new int [width * height * windowSize];
	columnX = new double [width * height * windowSize];
	
	assert(columnP != NULL);
	assert(columnI != NULL);
	assert(columnX != NULL);
	
	int result;
	
	/*Do the conversion with UMFPack*/
	result = umfpack_di_triplet_to_col(width * height,
				width * height,
				n,
				tripletI,
				tripletJ,
				tripletX,
				columnP,
				columnI,
				columnX, NULL);
	
	
	cout << "Matrix conversion returned " << result << endl;
	cout.flush();
	
	/*Free our old triplet matrix*/
	delete [] tripletI;
	delete [] tripletJ;
	delete [] tripletX;
	
	/*Reset pointers*/
	tripletI = (int*)NULL;
	tripletJ = (int*)NULL;
	tripletX = (double*)NULL;
	
	/*Officially change our objects state*/
	isTriplet = 0;
	
	return;
}
